-------------------------------------------------------------------------------
--
-- Projekt         I2C SLAVE
--
-- Compiler        ALTERA ModelSim 5.8e
--                 QUARTUS II 5.0
--
-- Autor           FPGA-USER
--
-- Datum           26/08/2005
--
-- Sprache         VHDL 93
--
-- Info            Testbench fuer I2C-Slave
--
-- Historie         
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use work.i2c_sim_pkg.all;
use work.i2c_syn_pkg.all;

entity tb is
end;

architecture tb of tb is
   
   ---------------------------------------------------
   -- I2C SLAVE
   ---------------------------------------------------
   component i2c_slave
	   generic
      (
         SDA_DELAY      : integer range 1 to 16 := 5
      );
      port
      (
         clk            : in std_logic;
         reset          : in std_logic;

         -- I2C Takt und Daten (SDA ist open Drain)
         scl            : in std_logic;
         sda            : inout std_logic;

         -- I2C Slave-Adresse, 7 Bit
         slv_adr        : in SLV_ADR_TYPE;
         sniffer_on     : in boolean;

         -- TX-Buffer-Signale 
         tx_data        : in BYTE;
         tx_wr          : in std_logic;
         tx_empty       : out std_logic;

         -- RX-Buffer-Signale
         rx_data        : out BYTE;
         rx_vld         : buffer std_logic;
         rx_ack         : in std_logic;

         -- Slave-Status
         busy           : out std_logic;
         error          : out I2C_SLV_ERROR_TYPE
      );
   end component;
   
   -- hier feste Slave-Adresse einstellen,
   -- kann weiter unten noch geandert werden
   signal SLAVE_ADDRESS : SLV_ADR_TYPE := 1;
   
   -- CLOCK / RESET   
   signal clk            : std_logic;
   signal rst_n          : std_logic;
   signal reset          : std_logic;

   -- I2C
   signal scl            : std_logic;
   signal sda            : std_logic;
   
   -- SLAVE-INTERFACE-SIGNALE
   signal slv_tx_data    : std_logic_vector(7 downto 0);
   signal slv_rx_data    : std_logic_vector(7 downto 0);
   signal slv_rx_vld     : std_logic;
   signal slv_rx_ack     : std_logic;
   signal slv_tx_wr      : std_logic;
   signal slv_tx_empty   : std_logic;

   signal slv_busy       : std_logic;
   signal slv_error      : I2C_SLV_ERROR_TYPE;
   
   signal cnt            : integer range 0 to 127;
   signal sniffer_on     : boolean := false;
   

begin
   
   ---------------------------------------------------------------
   -- SIMULATION I2C-SLAVE
   ---------------------------------------------------------------
   
   process
      variable ack : std_logic;
   begin
      sniffer_on <= false;
      wait until reset ='0';
      wait for 100 ns;

      -------------------------------------------------------
      report ("Test 1 : falsche Slave-Adresse verwenden");
      
      SLAVE_ADDRESS <= 1; -- Slave hat Adresse 1 !
      
      start_cond(scl, sda);
      send_slave_adr(scl, sda, 2, WRITE_DATA, ack); -- mit 2 ansprechen
      stop_cond(scl, sda);
      
      assert ack = NOT_ACK
         report "Slave hat trotz falscher Adresse Ack gesetzt"
            severity error;
            
      -------------------------------------------------------
      report ("Test 2 : 2 Bytes zum Slave senden");
      -- die Daten muessen am RX-Interface zu sehen sein
      -- (werden automatisch abgeholt)
      
      SLAVE_ADDRESS <= 127; -- mal ne andere 
      
      start_cond(scl, sda);
      send_slave_adr(scl, sda, 127, WRITE_DATA, ack);

      send_byte(scl, sda, x"5A");
      check_ack(scl, sda, ack);

      send_byte(scl, sda, x"3D");
      check_ack(scl, sda, ack);

      stop_cond(scl, sda); -- Transfer Ende
      
      -------------------------------------------------------
      report ("Test 3 : Sniffer Mode");
      sniffer_on <= true;
      -- Slave darf jetzt kein ACK senden
      -- es sollte jedes Byte vom Master mitgelesen werden,
      -- das Ansprechen des Slave mit Adr 127 entspricht dem
      -- Byte 0xFE
      -- die Daten werden wieder am RX-Interface automatisch gelesen
      wait for 1000 ns;
      
      start_cond(scl, sda);
      send_slave_adr(scl, sda, 127, WRITE_DATA, ack);
      
      send_byte(scl, sda, x"4F");
      check_ack(scl, sda, ack);
      
      stop_cond(scl, sda);
      
      
      
      wait;
   end process;
   
   ---------------------------------------------------------------
   
   -- empfangene Daten vom Slave automatisch abholen
   process(clk)
   begin
      if rising_edge(clk) then
         if slv_rx_vld='1' then
            slv_rx_ack <= '1';
         else
            slv_rx_ack <= '0';
         end if;
      end if;
   end process;
   
   -- Slave TX-Daten bereitstellen
   process
   begin
      wait until reset='0';
      loop
         wait until rising_edge(clk);
         
         if slv_tx_empty = '1' then
            slv_tx_wr <= '1';
            cnt <= (cnt + 1) mod 256;
            wait until rising_edge(clk);
            slv_tx_wr <= '0';
            wait until rising_edge(clk);
         end if;

      end loop;
   end process;
   slv_tx_data <= conv_std_logic_vector(cnt, slv_tx_data'LENGTH);

         
   -- FPGA-Clock erzeugen
   GEN_CLK : process
   begin
      loop
         clk <= '0';
         wait for 50 ns;
         clk <= '1';
         wait for 50 ns;
      end loop;
   end process;
   
   -- Asynchronen Reset erzeugen, wahlweise Hi / Lo-aktiv
   reset <= '1', '0' after 100 ns;
   rst_n <= not reset;

   i2c_slave_1 : i2c_slave
	   generic map
      (
         SDA_DELAY => 3
      )
	   port map
      (
         clk            => clk,
         reset          => reset,

         scl            => scl,
         sda            => sda,

         slv_adr        => SLAVE_ADDRESS,
         sniffer_on     => sniffer_on,

         tx_data        => slv_tx_data,
         tx_wr          => slv_tx_wr,
         tx_empty       => slv_tx_empty, 

         rx_data        => slv_rx_data,
         rx_vld         => slv_rx_vld,
         rx_ack         => slv_rx_ack,

         busy           => slv_busy,
         error          => slv_error
      );
   
   scl <= 'H'; -- immer weak high
   sda <= 'H';
end;